#!/usr/bin/bash

### See Conventions, Notes for purpose of open bracket at end of this line     {
### There should be nothing before this except the code to cause this script
### to be executed under bash.

### SECTION: Conventions, Notes ###############################################{

### - You will note comments of the form #{, #}{ and #}. These are simply to
###   allow bracket matching in editors that support it (such as vi) to make
###   it easier to navigate blocks of text. If it is necessary to introduce
###   extra { or } characters into the script for whatever reason (such as was
###   done in this very comment!), a matching bracket should be introduced as
###   well. You will note there is an open bracket near the beginning of this
###   script, and a close bracket as the last line. These should always match.
###   If not, there is an extraneous bracket in the script.
###
### - Related to the preceding bullet, sections have a name and have a matching
###   set of brackets at the start and end of the section. Search for SECTION
###   (all uppercase) in the text.
###
### - Comments are always introduced by three leading hash signs. This is to
###   differentiate hash used in the script for commenting out code
###   sections temporarily from true comments, and also to aid searching
###   for hash characters used in the script for purposes other than comments,
###   such as the #{ and #} constructs mentioned above.
###
### - There are a number of variable naming rules (conventions) put in place
###   for:
###   - readability
###   - to differentiate script variables from environment variables
###   - to designate variables as global, local, function arguments,
###     function arguments representing passback (volatile) variables.
###
###   The rules are outlined here:
###
###   - Local variable names are preceded with '_' (underscore) so a local
###     variable for storing a birthday might be _birthday
###
###   - An extension of the '_' prefix rule for local variables is that
###     function "parameters" are prepended with '_arg_'. The reason parameters
###     is in quotes in the preceding sentence is that technically, function
###     parameters in bash script are named $1, $2, etc., but we copy those
###     values to variables at the beginning of the functions for readability
###     later in the functions. It's these variables that use the '_arg_'
###     prefix rule.
###
###   - Global variable names are preceded with 'g_' so a global variable
###     for storing a verbosity flag might be g_verbosity.
###
###   - Variables are made read-only (constant) when possible. This is done
###     with the 'declare -r' or 'local -r' construct. Global read only
###     variables have names that are all uppercase, e.g., g_NULL.
###
###   - For variables that will be used to hold integer only values, they are
###     defined as such using the 'declare -i' or 'local -i' construct. There's
###     a catch to this though; when comparing integers for anything other than
###     equality, you must use the '-ge', '-gt', '-le' or '-lt' comparators and
###     not the '>' or '<' comparators. The latter will cause the comparison to
###     be done on the string value and not the integer value, so '2 > 17' would
###     be true while '2 -gt 17' would be false.
###
### - Functions names are preceded with 'fn_', so a function for clearing the
###   screen might be called 'fn_clearscreen'. You will also note that the open
###   bracket { and the close bracket } appear in column zero, i.e., as the
###   first character of the line. This is so the '[[' shortcut in vi will
###   work to find the beginning of the function you are in.
###
### - To differentiate '>' meaning 'greater than' (lexical comparison)
###   from '>' meaning 'redirect output to file', and to aid searching for
###   either explicitly, the former meaning will have a leading and trailing
###   space and the latter meaning (and its derivatives) will be immediately
###   followed by the file name with no intervening space. The same thing
###   will be done for '<' ('less than' or 'read from file') and its
###   derivatives.
###
### - Script lines are limited to a maximum of 80 characters, where possible,
###   to aid readability, which results in easier maintenance of the script
###   except where making a line meet this length limitation makes it LESS
###   readable in which case the rule is ignored as the overriding goal here
###   is readability.
###
### - The construct '&>filepath' is used to redirect standard error (file
###   descriptor 2) and standard out (file descriptor 1) to the file
###   "filepath". The '&' and '>' must appear together with no space between,
###   so as to not be interpreted as "run this command in the background and
###   redirect stdout". In other words 'ls &>/dev/null' is not the same as
###   'ls & >/dev/null'
###
### - The $(command) construct is used rather than the `command` construct
###   to execute commands for the following reasons:
###   - There is support for bracket matching in most editors (such as the
###     % command in vi.)
###   - You don't have to use escape like you do on the back quote when
###     nesting, e.g., you can say $(command1 $(command2)) rather than
###     `command1 \`command2\``
###   - You don't have to escape lexical comparators '<' and '>'.
###
### - IMPORTANT! bash has a nasty habit of allowing local variables to a
###   function to be inherited by children. i.e., if fn_a() declares local
###   variable '_a', and fn_a() calls fn_b(), then '_a' will be in scope
###   within fn_b() and changes to it will be propagated back to fn_a().
###   Worse, if fn_b() also declares local variable '_a', then any changes
###   to '_a' within fn_b() will not be propagated back to fn_a(), which is
###   expected, except that if we use parameter dereferencing to set '_a'
###   within fn_b() *by reference* we would expect that it would be changes
###   within fn_a(), but in fact it's not! This causes a problem when we pass
###   in a passback variable to a function, and the name of the variable
###   happens to be the same as the name of the temporary variable used to
###   store the value we are gathering to later modify in the dereference.
###   We handle this by either including the function name in temporary
###   variable names, or by dereferencing the passed value directly and
###   not using temporary variables.
###
### - We use the [[ ]] test construct rather than the more common [ ]. The
###   main reasons we do this are:
###   - It has syntax and parameter checking, so you don't need to use quotes,
###     nor do you need to do things like comparing "X$var" to "Xval", you can
###     instead compare $var to val without any problems. This may seem trivial
###     at first, until you consider that the old style does not do this, and
###     you must always remember to protect the tests lest your code fail when
###     a value is null.
###   - Related to the preceding point, you can do simple existence tests by
###     saying [[ $var ]] rather than the less readable [ -n "X$var" ]
###   - It supports || and && within the test so it's more readable than the
###     -o and -a constructs.

### END-SECTION: Conventions, Notes ###########################################}


### SECTION: Global Variables #################################################{

### To differentiate script variables from environment variables, global
### variable names are preceded with 'g_' so a local variable for storing a
### birthday might be _birthday, and a global variable for storing a verbosity
### flag might be g_verbosity.

declare -r g_SCRIPTDIR=/nn/patches/scripts
declare -r g_PROGNAME=$(basename $0)
declare -r g_LOGFILE=/var/log/post_patch.log

### The scripts are named with a specific file naming format, specifically,
### MSIIP.sh, where:
### M:  The master prefix, i.e., the prefix all the scripts have, kept
###     in g_PFX_MASTER
### S:  The sub-prefix:
###     - g_SUBPFX_PRE_REBOOT:    These scripts are executed prior to the
###                               reboot script and any post reboot scripts.
###     - g_SUBPFX_REBOOT:        If this script exists, it is executed. No
###                               other scripts will be run after this point.
###     - g_SUBPFX_STANDARD:      These scripts are run last, and only of
###                               there was not a reboot script.
### II: A 2 digit number indicating the serial order in which to run
###     the scripts. The sort precedence of these values comes after
###     the sub-prefix. These values need not be unique amongst scripts
###     whose execution order is independent. This serial order is not
###     required for the reboot scripts, i.e., the scripts whose sub-prefix
###     is g_SUBPFX_REBOOT.
### P:  The name of the package. Not required for the reboot scripts, i.e.,
###     the scripts whose sub-prefix is g_SUBPFX_REBOOT.
### Note also that for the reboot scripts, i.e., the scripts whose sub-prefix
### is g_SUBPFX_REBOOT, additional characters may appear after the sub-prefix.
### This introduces the capability of having more than one reboot script, but
### allowing control of the call order should that capability be needed.
declare -r g_PFX_MASTER=action_
declare -r g_SUBPFX_PRE_REBOOT=preReboot
declare -r g_SUBPFX_REBOOT=reboot
declare -r g_SUBPFX_STANDARD=standard

### When we want to execute any reboot script, we rename it. The reason we
### rename it is because we will need to remove it, but in the event we don't
### have time to remove it prior to the reboot happening, we at least want to
### be sure it doesn't get executed again by a subsequent call to this script.
### The variable g_TMP_REBOOT_FILE holds the name we rename the reboot file to.
declare -r g_TMP_REBOOT_FILE=$g_SCRIPTDIR/tmpreboot.sh

### END-SECTION: Global Variables #############################################}


### SECTION: Script setup #####################################################{

### This script must be run as root.
if [[ $UID != 0 ]]
then #{
    echo "You must run this script as root" | tee -a $g_LOGFILE >&2
    exit 1
fi #}

### This script should be made executable and run directly rather than
### sourced from the shell (i.e., it should be run using the command line:
###   ./scriptname
### rather than:
###   . ./scriptname
### or:
###   bash ./scriptname
### Sourcing the script from the shell breaks the behaviour of the exit
### command so we will now check to be sure the script was run directly.
if [[ ${-/i/} != $- ]]
then #{
    echo "You must not source this script, but run it directly instead" | tee -a $g_LOGFILE >&2
    exit 1
fi #}

### END-SECTION: Script setup #################################################}


### SECTION: Function definitions #############################################{

### Args: $1 Path name of script to run.
###       $2 Set to YES if this is a reboot script.
fn_runscript()
{
    if [[ $# -ne 2 ]]
    then #{
        echo "Function $FUNCNAME takes two arguments" | tee -a $g_LOGFILE >&2
        exit 1
    fi #}

    local    _arg_rs_script=$1
    local -r _arg_rs_REBOOT=$2

    if [[ $_arg_rs_REBOOT != YES && $_arg_rs_REBOOT != NO ]]
    then #{
        echo "_arg_rs_REBOOT must be \"YES\" or \"NO\", but is \"$_arg_rs_REBOOT\"" | tee -a $g_LOGFILE >&2
        exit 1
    fi #}

    echo "Found $_arg_rs_script" >>$g_LOGFILE

    ### If we're about to execute a reboot script, we have some
    ### additional tasks.
    if [[ $_arg_rs_REBOOT == YES ]]
    then #{

        ### Remove any existing standard scripts.
        rm -f $g_SCRIPTDIR/$g_PFX_MASTER$g_SUBPFX_STANDARD[0-9][0-9]*\.sh >&/dev/null

        ### Remove any old temporary reboot files, if any.
        rm -f $g_TMP_REBOOT_FILE >&/dev/null

        ### Rename the reboot file to our temporary file. The reason we rename
        ### it is twofold:
        ### 1) We will need to remove it after it's executed, but in the event
        ###    we don't have time to remove it prior to the reboot happening,
        ###    we at least want to be sure it doesn't get executed again by a
        ###    subsequent call to this script.
        ### 2) Renaming the file allows us to generically remove the remaining
        ###    reboot files.
        mv -f $_arg_rs_script $g_TMP_REBOOT_FILE >&/dev/null

        ### Remove any remaining reboot scripts.
        rm -f $g_SCRIPTDIR/$g_PFX_MASTER$g_SUBPFX_REBOOT*\.sh >&/dev/null

        ### Let's modify the argument indicating the script we're working on to
        ### point to the temporary reboot file.
        _arg_rs_script=$g_TMP_REBOOT_FILE
    fi #}

    ### Make the script executable.
    chmod u+x $_arg_rs_script >&/dev/null

    if [[ -x $_arg_rs_script ]]
    then #{
        echo "Executing script $_arg_rs_script" >>$g_LOGFILE
        $_arg_rs_script >>$g_LOGFILE 2>&1
    else #}{
        echo "Script $_arg_rs_script not executable" >>$g_LOGFILE
    fi #}

    rm -f $_arg_rs_script >&/dev/null

    echo >>$g_LOGFILE
}

### END-SECTION: Function definitions #########################################}


### SECTION: MAIN #############################################################{

### So that the main loop can use local variables, we encompass the entire
### main block of code in fn_main().
fn_main()
{
    echo "$g_PROGNAME begin $(date)" >>$g_LOGFILE

    local _main_script=

    ### Run all the pre-reboot scripts, if any.
    for _main_script in $(ls $g_SCRIPTDIR/$g_PFX_MASTER$g_SUBPFX_PRE_REBOOT[0-9][0-9]*\.sh 2>/dev/null)
    do #{
        fn_runscript $_main_script NO
    done #}

    ### Run the "first" pre-reboot script, if any.
    read _main_script < <(ls $g_SCRIPTDIR/$g_PFX_MASTER$g_SUBPFX_REBOOT*\.sh 2>/dev/null)
    if [[ $_main_script ]]
    then #{

        # BFinch May 22 2007 - need to update status for Element Manager and wait for EM to read it:
        echo "Apply_Reboot" >/nn/patches/info/active.status
        sleep 3

        fn_runscript $_main_script YES

        ### We just scheduled a reboot, let's terminate the script.
        exit 0
    fi #}

    ### Run all the standard scripts, if any. Note that if there were any
    ### reboot scripts, the standard scripts would have been removed.
    for _main_script in $(ls $g_SCRIPTDIR/$g_PFX_MASTER$g_SUBPFX_STANDARD[0-9][0-9]*\.sh 2>/dev/null)
    do #{
        # BFinch May 23 2007 - need to update status for Element Manager 
        echo "Apply_Success" >/nn/patches/info/active.status

        fn_runscript $_main_script NO
    done #}

    echo "$g_PROGNAME done $(date)" >>$g_LOGFILE

    exit 0
}

### Call the main loop.
fn_main $*

exit 0

### SECTION: MAIN #############################################################}


### There should be nothing after the next line.
### See Conventions, Notes for purpose of close bracket at end of this line    }
